import asyncio
import hashlib
import re
from collections import deque
from dataclasses import dataclass
from enum import Enum, auto
from typing import List, Literal, Optional, Protocol, Set, Tuple, Union

import aiohttp
from fastapi import WebSocket
from lonelypsp.stateful.message import S2B_Message
from lonelypsp.stateful.messages.confirm_receive import S2B_ConfirmReceive
from lonelypsp.stateful.messages.continue_receive import S2B_ContinueReceive
from lonelypsp.stateful.messages.notify_stream import (
    S2B_NotifyStreamStartCompressed,
    S2B_NotifyStreamStartUncompressed,
)

from lonelypss.config.config import Config
from lonelypss.util.sync_io import SyncIOBaseLikeIO, SyncReadableBytesIO
from lonelypss.util.websocket_message import WSMessage
from lonelypss.util.ws_receiver import BaseWSReceiver, FanoutWSReceiver

try:
    import zstandard
except ImportError:
    ...


@dataclass
class ConnectionConfiguration:
    """Describes what settings were negotiated at the connection level.
    These are requested by the CONFIGURE packet by the subscriber, then
    we reduce it to what we are willing to do and do _not_ tell the
    subscriber what we chose (for convenience).
    """

    enable_zstd: bool
    """True if the subscriber can use zstandard-compressed data and might send
    zstandard-compressed data, false if the broadcaster should not send
    zstandard-compressed data and should reject zstandard-compressed data.
    """

    enable_training: bool
    """True if the subscriber can receive custom compression dictionaries, so
    it may make sense for the broadcaster to train one and send it. False if the
    subscriber will not accept custom dictionaries and so the broadcaster
    shouldn't waste time training one.
    """


class StateType(Enum):
    """Discriminator value for the state the websocket is in."""

    ACCEPTING = auto()
    """The websocket was just initialized but we are waiting to finish the
    websocket-level handshake (websocket.accept())
    """

    WAITING_CONFIGURE = auto()
    """The websocket-level handshake is finished and we are waiting for the
    subscriber to send a CONFIGURE message
    """

    OPEN = auto()
    """The main state of the websocket, where we are sending and receiving
    messages.
    """

    CLOSING = auto()
    """We are performing final cleanup tasks before closing the websocket."""

    CLOSED = auto()
    """The websocket is closed and no longer usable."""


@dataclass
class StateAccepting:
    """The variables available when in the ACCEPTING state"""

    type: Literal[StateType.ACCEPTING]
    """discriminator value"""

    websocket: WebSocket
    """the underlying socket that we will communciate over"""

    broadcaster_config: Config
    """the configuration of the broadcaster, which will be used to e.g.
    check authorization headers, load preset compression dictionaries, etc
    """

    internal_receiver: FanoutWSReceiver
    """how the broadcaster can receive messages sent from subscribers to
    other broadcasters
    """


class InternalMessageType(Enum):
    """The types of messages that we can receive from the internal receiver.
    These are pushed immediately into a queue so they can be processed by the
    standard coroutine as soon as possible while otherwise being unaltered.
    """

    SMALL = auto()
    """A messages small enough that we can hold it in memory"""

    LARGE = auto()
    """A message which may be large enough that we may not want to hold it
    in memory.
    """

    MISSED = auto()
    """A message which indicates this broadcaster may have missed a message
    from another broadcaster on a relevant topic
    """


@dataclass
class InternalSmallMessage:
    """A message from the receiver that is small enough to hold in memory"""

    type: Literal[InternalMessageType.SMALL]
    """discriminator value"""

    topic: bytes
    """the topic the message was sent to"""

    data: bytes
    """the uncompressed message data"""

    sha512: bytes
    """a trusted sha512 hash of the data"""


@dataclass
class InternalLargeMessage:
    """A message from the receiver that may be too large to hold in memory"""

    type: Literal[InternalMessageType.LARGE]
    """discriminator value"""

    stream: SyncReadableBytesIO
    """a readable stream of the message. this stream is not tellable, seekable, or
    closeable; use the finished event to indicate that you will no longer be
    accessing the stream.

    the read(n) function may read past the end of the message if n is too large;
    you must instead be careful to read only the number of bytes you expect.
    """

    length: int
    """the length of the message body"""

    finished: asyncio.Event
    """must be set as soon as the stream will no longer be accessed for this
    websocket connection; allows the handle to be used by other connections
    """

    topic: bytes
    """the topic the message was sent to"""

    sha512: bytes
    """a trusted sha512 hash of the message body"""


@dataclass
class InternalMissedMessage:
    """A message from the receiver that indicates this broadcaster may have missed
    a message from another broadcaster on a relevant topic
    """

    type: Literal[InternalMessageType.MISSED]
    """discriminator value"""

    topic: bytes
    """the topic the message was sent to"""


InternalMessage = Union[
    InternalSmallMessage, InternalLargeMessage, InternalMissedMessage
]


class WaitingInternalMessageType(Enum):
    """Describes the types of messages that we have pulled off the internal
    receiver queue but were not able to process as the websocket was busy.

    This is similar to the internal messages, but we augmented them by assigning
    them an identifier, and switched the data structure they are stored in since
    we no longer have to share it with other asyncio tasks.
    """

    SPOOLED_LARGE = auto()
    """A message that may be large enough that we may not want to hold it in
    memory. Unlike with the original internal message, when we realized we
    weren't able to process it immediately we copied the data to another location
    and released the original stream, so this can be closed with a typical close()
    call.
    """


@dataclass
class WaitingInternalSpooledLargeMessage:
    """A message from the receiver that may be too large to hold in memory, after
    we've copied it to a new location and released the original stream.
    """

    type: Literal[WaitingInternalMessageType.SPOOLED_LARGE]
    """discriminator value"""

    stream: SyncIOBaseLikeIO
    """a readable, seekable, tellable, and closeable stream to the copied data. closing
    this stream deletes our copy of the data

    0 is the start of the message and does not contain any additional data besides the
    message (ie., cannot read past)
    """

    length: int
    """the length of the message body in bytes"""

    topic: bytes
    """the topic the message was sent to"""

    sha512: bytes
    """a trusted sha512 hash of the message body"""


WaitingInternalMessage = Union[
    InternalSmallMessage, InternalMissedMessage, WaitingInternalSpooledLargeMessage
]


class SimplePendingSendType(Enum):
    """Describes tasks besides WaitingInternalMessage that will eventually need
    exclusive access to sending data through the websocket
    """

    PRE_FORMATTED = auto()
    """A message that is ready to be sent via send_bytes"""


@dataclass
class SimplePendingSendPreFormatted:
    """A message that is ready to be sent via send_bytes on the websocket once it's
    not busy
    """

    type: Literal[SimplePendingSendType.PRE_FORMATTED]
    """discriminator value"""

    data: bytes
    """the data to send"""


class AsyncioWSReceiver(BaseWSReceiver, Protocol):
    """The receiver type that we use for the websocket where all the function
    implementations just forward to exposed data structures in a relatively
    obvious way. This is used because the caller essentially wants to manipulate
    the functions behaviors directly and synchronously.
    """

    @property
    def exact_subscriptions(self) -> Set[bytes]:
        """The exact topics that the receiver returns True to from is_relevant,
        that the caller can mutate
        """

    @property
    def glob_subscriptions(self) -> List[Tuple[re.Pattern, str]]:
        """The glob patterns that the receiver returns True to from is_relevant,
        that the caller can mutate. The first element in each tuple is the regex
        that matches the pattern, and the second element is the original pattern
        """

    @property
    def queue(self) -> asyncio.Queue[InternalMessage]:
        """The queue that this receiver pushes messages to. For large messages,
        this involves generating a finished event that is wait()'d before returning
        """


@dataclass
class CompressorTrainingDataCollector:
    """Combines the related state for collecting training data to train a custom
    compression dictionary for this connection
    """

    messages: int
    """the number of messages that have been collected so far"""

    length: int
    """how much training data has been collected; the file length will be longer
    as it includes the length of each sample before the sample itself (4 bytes per sample)
    """

    tmpfile: SyncIOBaseLikeIO
    """a readable, writable, seekable, tellable, closable file-like object where we are storing 
    the data to train the dictionary. closing this file will delete the data

    see collector_utils for a more convenient interface

    typically, if writing a small message, the flow is:
    - seek to the end (`seek(0, os.SEEK_END)`)
    - write the length prefix (`length.to_bytes(4, "big")`)
    - write the data
    - yield to the event loop

    for writing a large message, the flow is:
    - store a reference to the collector as the states collector may change
    - create an asyncio.Event, add it to the 'pending' set
    - seek to the end (`seek(0, os.SEEK_END)`)
    - write the length prefix (`length.to_bytes(4, "big")`)
    - remember the current position (`pos = tell()`)
    - write zeros for the data
    - yield to the event loop
    - when you receive more data:
      - seek to the position (`seek(pos, os.SEEK_SET)`)
      - write the data
      - remember the new position (`pos += len(data)`)
      - repeat until all data is written
    - mark the event finished
    - remove the event from the pending set
    """

    pending: Set[asyncio.Event]
    """while this set is non-empty the data in the tmpfile may be in a partial
    state; when this is empty, the data in the tmpfile is complete and can be
    used to train a dictionary
    """


class CompressorTrainingInfoType(Enum):
    """If training is allowed, the potential states we are in with regard to training a
    compression dictionary
    """

    BEFORE_LOW_WATERMARK = auto()
    """We are waiting for data to build a new dictionary using the low watermark settings"""
    BEFORE_HIGH_WATERMARK = auto()
    """We are waiting for data to build a new dictionary using the high watermark settings"""
    WAITING_TO_REFRESH = auto()
    """We built a dictionary recently; once some time passes, we'll build another one"""


@dataclass
class CompressorTrainingInfoBeforeLowWatermark:
    """We are waiting for data to build a new dictionary using the low watermark settings"""

    type: Literal[CompressorTrainingInfoType.BEFORE_LOW_WATERMARK]
    """discriminator value"""

    collector: CompressorTrainingDataCollector
    """The collector where the data is stored"""

    compressor_id: int
    """The compressor id we will use for the next dictionary once we have
    produced it
    """


@dataclass
class CompressorTrainingInfoBeforeHighWatermark:
    """We are collecting data to build a new dictionary using the high watermark settings"""

    type: Literal[CompressorTrainingInfoType.BEFORE_HIGH_WATERMARK]
    """discriminator value"""

    collector: CompressorTrainingDataCollector
    """The collector where the data is stored"""

    compressor_id: int
    """The compressor id we will use for the next dictionary once we have
    produced it
    """


@dataclass
class CompressorTrainingInfoWaitingToRefresh:
    """We built a dictionary recently; once some time passes, we'll build another one"""

    type: Literal[CompressorTrainingInfoType.WAITING_TO_REFRESH]
    """discriminator value"""

    last_built: float
    """the time the last high watermark dictionary was built in seconds from the epoch
    (as if by `time.time()`)
    """

    compressor_id: int
    """The compressor id we will use for the next dictionary once we have
    produced it
    """


CompressorTrainingInfo = Union[
    CompressorTrainingInfoBeforeLowWatermark,
    CompressorTrainingInfoBeforeHighWatermark,
    CompressorTrainingInfoWaitingToRefresh,
]


class CompressorState(Enum):
    """Describes the state that a zstandard compressor is in"""

    PREPARING = auto()
    """We are in the process of preparing the compressor for use"""

    READY = auto()
    """The compressor is ready to use right now; the individual zstandard compressors/
    decompressors are not async or thread safe, but the pool of them is async safe
    """


@dataclass
class CompressorReady:
    """A compressor which is ready to use"""

    type: Literal[CompressorState.READY]
    """discriminator value"""

    identifier: int
    """the integer identifier for this compressor; positive integers only."""

    level: int
    """the compression level the compressor is set to"""

    min_size: int
    """the minimum size, in bytes, inclusive, that a message can be for the broadcaster
    to choose this compressor for the message. 0 means no minimum size. Note that the
    subscriber may use this compressor for messages smaller than this size and the
    broadcaster will still decompress it.
    """

    max_size: Optional[int]
    """the maximum size, if any, in bytes, exclusive, that a message can be for the broadcaster
    to choose this compressor for the message. None means no maximum size. Note that the
    subscriber may use this compressor for messages larger than this size and the
    broadcaster will still decompress it.
    """

    data: "Optional[zstandard.ZstdCompressionDict]"
    """if there is a custom compression dictionary, that dictionary, otherwise None"""

    compressors: "List[zstandard.ZstdCompressor]"
    """the zstandard compressor objects that are not in use. pulled LIFO as it is
    preferable to reuse the same object as much as possible.
    
    WARN: individual compressors are not asyncio safe
    """

    decompressors: "List[zstandard.ZstdDecompressor]"
    """the zstandard decompressors that are not in use. pulled LIFO as it is
    preferable to reuse the same object as much as possible.
    
    WARN: decompressors are not asyncio-safe
    """


@dataclass
class CompressorPreparing:
    """A compressor that isn't ready to use yet"""

    type: Literal[CompressorState.PREPARING]
    """discriminator value"""

    identifier: int
    """the integer identifier for this compressor; positive integers only.

    1 is reserved for the compressor with no special compression dictionary,
    i.e., data is compressed by building a compression dictionary for each
    message. this works best on moderate sized messages (e.g., at least 16kb)
    as otherwise the overhead of the dictionary size may outweigh the benefits
    of using it

    2-65535 are reserved for preset compression dictionaries, where the compression
    dictionary is distributed out of band

    65536 and above are custom dictionaries built for the connection
    """

    task: asyncio.Task[CompressorReady]
    """The task that is working on preparing the compressor for use"""


Compressor = Union[CompressorReady, CompressorPreparing]


@dataclass
class NotifyStreamState:
    """Keeps track of the combined state of the last related NOTIFY_STREAM messages
    from the subscriber that have been processed.
    """

    identifier: bytes
    """the message identifier that we are receiving, chosen arbitrarily by the
    subscriber
    """

    first: Union[S2B_NotifyStreamStartUncompressed, S2B_NotifyStreamStartCompressed]
    """The first stream message with this id, with the payload stripped out"""

    part_id: int
    """The last part id that we received"""

    body_hasher: "hashlib._Hash"
    """the hash object that is producing the sha512 hash of the body as it comes in"""

    body: SyncIOBaseLikeIO
    """a writable, seekable, tellable, closeable file-like object where we are storing
    the body of the message as it comes in. closing this file will delete the data
    """


@dataclass
class StateWaitingConfigure:
    """The variables available in the WAITING_CONFIGURE state, before the subscriber
    has sent the CONFIGURE message
    """

    type: Literal[StateType.WAITING_CONFIGURE]
    """discriminator value"""

    websocket: WebSocket
    """the underlying socket that we will communciate over. We exclusively read
    bytes via the read_task and do not send bytes in this state
    """

    broadcaster_config: Config
    """configures how the broadcaster acts and what functionality it supports"""

    internal_receiver: FanoutWSReceiver
    """how the broadcaster can receive messages sent from subscribers to
    other broadcasters
    """

    read_task: asyncio.Task[WSMessage]
    """The task that is reading the next message from the websocket, which
    ought to be the configure message
    """


@dataclass
class StateOpen:
    """The variables available when in the OPEN state, which comprises the majority
    of the time the websocket connection is active
    """

    type: Literal[StateType.OPEN]
    """discriminator value"""

    websocket: WebSocket
    """the underlying socket that we will communciate over.
    
    we exclusively send bytes to this websocket in the send_task, and we 
    exclusively read bytes from this websocket in the read_task
    """

    broadcaster_config: Config
    """configures how the broadcaster acts and what functionality it supports"""

    connection_config: ConnectionConfiguration
    """the negotiated configuration taking into account the clients CONFIGURE message"""

    nonce_b64: str
    """The base64 encoded connection nonce, which is the format we need it in
    for producing URLs
    """

    internal_receiver: FanoutWSReceiver
    """allows the broadcaster to receive notifications sent to other broadcasters"""

    my_receiver: AsyncioWSReceiver
    """the receiver we registered with the fanout receiver"""

    my_receiver_id: int
    """the id we received when registering `my_receiver` with `internal_receiver`"""

    client_session: aiohttp.ClientSession
    """the aiohttp ClientSession for notifying other subscribers when this subscriber
    notifies the broadcaster via NOTIFY or NOTIFY_STREAM
    """

    compressors: List[Compressor]
    """The compressors that the broadcaster may try to use or will accept,
    in the order they were initialized.
    
    When rotating in a new compressor, generally the oldest (closest to the
    front) is removed, excluding the first entry (identifier 1).

    May be empty is compresion is disabled in the broadcaster config or the
    connection config
    """

    compressor_training_info: Optional[CompressorTrainingInfo]
    """If we may potentially collect data for training a new compression dictionary,
    based on the broadcaster config and the connection config, contains the state
    related to that
    """

    broadcaster_counter: int
    """Whenever the broadcaster produces authorization headers a URI identifying
    the subscriber is required. For the HTTP api, this is the actual URL that
    the broadcaster is reaching out to. For the websocket api, instead, this is
    a unique value that is comprised of the connection nonce and this counter,
    where this counter is incremented after each use.

    Starts at 1
    """

    subscriber_counter: int
    """Whenever the subscriber produces authorization headers a URI identifying
    the subscriber is required. For the HTTP api, this is the actual URL the
    broadcaster can reach the subscriber. For the websocket api, instead, this is
    a unique value that is comprised of the connection nonce and this counter,
    where this counter is decremented after each use.

    Starts at -1
    """

    read_task: asyncio.Task[WSMessage]
    """The task that is currently responsible for getting the next message on the
    websocket from the ASGI server
    """

    internal_message_task: asyncio.Task[InternalMessage]
    """The task that is currently responsible for getting the next message from 
    my_receiver.queue
    """

    notify_stream_state: Optional[NotifyStreamState]
    """If there is a NOTIFY_STREAM sequence that has not completed yet, the state
    of that sequence, otherwise None
    """

    send_task: Optional[asyncio.Task[None]]
    """If a task is currently responsible for sending data over the websocket, that
    task, otherwise None to indicate nothing is sending data over the websocket
    """

    process_task: Optional[asyncio.Task[None]]
    """If a task is currently processing one of the incoming messages, the task that
    is doing that, otherwise None
    """

    unprocessed_messages: deque[S2B_Message]
    """If the broadcaster has received messages from the subscriber that it has not
    yet processed, the messages that have not been processed in the order they were 
    received.

    basic idea: websocket incoming -> read_task -> unprocessed_messages -> process_task
    """

    unsent_messages: deque[Union[WaitingInternalMessage, SimplePendingSendPreFormatted]]
    """If the broadcaster knows it needs to send messages to the subscriber that it
    has not yet sent because the websocket was busy (send_task was set), the messages
    that need to be sent in the order they should be sent.
    
    Note that for internal messages, we don't format them until we have exclusive access
    to sending messages over the websocket for efficiencies sake.

    - scenario 1: process_task -> unsent_messages -> send_task (e.g., confirm notify)
    - scenario 2: my_receiver.queue -> message_task -> unsent_messages -> send_task (e.g., receive)
    """

    expecting_acks: asyncio.Queue[Union[S2B_ContinueReceive, S2B_ConfirmReceive]]
    """When the broadcaster sends a RECEIVE_STREAM message to the subscriber the
    subscriber must acknowledge the message. If there are enough messages that have
    not been acknowledged, the broadcaster begins to wait for acknowledgements before
    sending more messages, which can cause unsent_messages to grow to its limit which
    will then have the websocket disconnected by the broadcaster

    The left of this queue is the next ack we expect to receive; if we receive any
    other ack there was a problem
    """

    backgrounded: Set[asyncio.Task[None]]
    """A set of tasks that need some time to complete, and should be checked to make
    sure they did not error (otherwise, the broadcaster should disconnect the subscriber),
    but do not have any other particular influence
    """


@dataclass
class StateClosing:
    """The variables available when in the CLOSING state, where we are closing
    the websocket
    """

    type: Literal[StateType.CLOSING]
    """discriminator value"""

    websocket: WebSocket
    """the websocket we are waiting to close"""

    exception: Optional[BaseException] = None
    """the exception we intend to raise once cleanup is complete, if any"""


@dataclass
class StateClosed:
    """The variables available when in the CLOSED state, where the websocket is
    closed and no longer usable
    """

    type: Literal[StateType.CLOSED]
    """discriminator value"""


State = Union[
    StateAccepting, StateWaitingConfigure, StateOpen, StateClosing, StateClosed
]



import asyncio
from typing import TYPE_CHECKING

from lonelypss.ws.handlers.protocol import StateHandler
from lonelypss.ws.state import State, StateClosing, StateType, StateWaitingConfigure
from lonelypss.ws.util import make_websocket_read_task


async def handle_accepting(state: State) -> State:
    """Accepts the websocket connection within the timeout period then moves to
    waiting for the configure message
    """

    assert state.type == StateType.ACCEPTING
    try:
        await asyncio.wait_for(
            state.websocket.accept(),
            timeout=state.broadcaster_config.websocket_accept_timeout,
        )
    except asyncio.TimeoutError:
        return StateClosing(type=StateType.CLOSING, websocket=state.websocket)

    return StateWaitingConfigure(
        type=StateType.WAITING_CONFIGURE,
        websocket=state.websocket,
        broadcaster_config=state.broadcaster_config,
        internal_receiver=state.internal_receiver,
        read_task=make_websocket_read_task(state.websocket),
    )


if TYPE_CHECKING:
    _: StateHandler = handle_accepting



import asyncio
from collections import deque
from typing import TYPE_CHECKING, Iterable, List, Optional, SupportsIndex, Union, cast

from lonelypss.ws.handlers.open.check_background_tasks import check_background_tasks
from lonelypss.ws.handlers.open.check_compressors import check_compressors
from lonelypss.ws.handlers.open.check_internal_message_task import (
    check_internal_message_task,
)
from lonelypss.ws.handlers.open.check_process_task import check_process_task
from lonelypss.ws.handlers.open.check_read_task import check_read_task
from lonelypss.ws.handlers.open.check_result import CheckResult
from lonelypss.ws.handlers.open.check_send_task import check_send_task
from lonelypss.ws.handlers.open.errors import (
    NormalDisconnectException,
    combine_multiple_exceptions,
)
from lonelypss.ws.handlers.open.processors.processor import process_any
from lonelypss.ws.handlers.protocol import StateHandler
from lonelypss.ws.state import (
    CompressorState,
    CompressorTrainingInfoType,
    SimplePendingSendPreFormatted,
    State,
    StateClosing,
    StateOpen,
    StateType,
    WaitingInternalMessage,
    WaitingInternalMessageType,
)


async def handle_open(state: State) -> State:
    """Makes some progress, waiting if necessary, and returning the new state. This
    may be the same state reference, allowing the caller to manage the required looping.

    It is intended that this never raises exceptions
    """
    assert state.type == StateType.OPEN

    _disconnected_receiver = False
    try:
        try:
            if await check_send_task(state) == CheckResult.RESTART:
                return state

            if await check_internal_message_task(state) == CheckResult.RESTART:
                return state

            if await check_read_task(state) == CheckResult.RESTART:
                return state

            if await check_process_task(state) == CheckResult.RESTART:
                return state

            if await check_background_tasks(state) == CheckResult.RESTART:
                return state

            if await check_compressors(state) == CheckResult.RESTART:
                return state

            await asyncio.wait(
                [
                    *([state.send_task] if state.send_task is not None else []),
                    state.internal_message_task,
                    state.read_task,
                    *([state.process_task] if state.process_task is not None else []),
                    *state.backgrounded,
                    *[
                        compressor.task
                        for compressor in state.compressors
                        if compressor.type == CompressorState.PREPARING
                    ],
                ],
                return_when=asyncio.FIRST_COMPLETED,
            )
            return state
        except NormalDisconnectException:
            if state.send_task is not None:
                state.send_task.cancel()
                state.send_task = None

            state.send_task = cast(
                asyncio.Task[None], asyncio.create_task(asyncio.Event().wait())
            )
            old_unsent = state.unsent_messages
            state.unsent_messages = VoidingDeque()

            while old_unsent:
                _cleanup(old_unsent.popleft())

            state.internal_message_task.cancel()

            if not _disconnected_receiver:
                _disconnected_receiver = True
                await _disconnect_receiver(state)

            if state.process_task is not None:
                await state.process_task
                state.process_task = None

            while state.unprocessed_messages:
                await process_any(state, state.unprocessed_messages.popleft())

            raise
    except BaseException as cause_for_cleanup_exc:
        cleanup_exceptions: List[BaseException] = []

        for compressor in state.compressors:
            if compressor.type == CompressorState.PREPARING:
                compressor.task.cancel()

        if state.compressor_training_info is not None:
            if (
                state.compressor_training_info.type
                != CompressorTrainingInfoType.WAITING_TO_REFRESH
            ):
                try:
                    state.compressor_training_info.collector.tmpfile.close()
                except BaseException as e2:
                    cleanup_exceptions.append(e2)

        state.read_task.cancel()
        state.internal_message_task.cancel()

        if state.notify_stream_state is not None:
            try:
                state.notify_stream_state.body.close()
            except BaseException as e2:
                cleanup_exceptions.append(e2)

        if state.send_task is not None:
            state.send_task.cancel()

        if state.process_task is not None:
            state.process_task.cancel()

        for msg in state.unsent_messages:
            if msg.type == WaitingInternalMessageType.SPOOLED_LARGE:
                try:
                    msg.stream.close()
                except BaseException as e2:
                    cleanup_exceptions.append(e2)

        for task in state.backgrounded:
            task.cancel()

        if not _disconnected_receiver:
            _disconnected_receiver = True
            try:
                await _disconnect_receiver(state)
            except BaseException as e2:
                cleanup_exceptions.append(e2)

        try:
            await state.client_session.close()
        except BaseException as e2:
            cleanup_exceptions.append(e2)

        result_exception: Optional[BaseException] = None

        if cleanup_exceptions:
            result_exception = combine_multiple_exceptions(
                "cleaning up from open state", cleanup_exceptions
            )
            result_exception.__context__ = cause_for_cleanup_exc
        elif not isinstance(cause_for_cleanup_exc, NormalDisconnectException):
            result_exception = cause_for_cleanup_exc

        return StateClosing(
            type=StateType.CLOSING,
            websocket=state.websocket,
            exception=result_exception,
        )


if TYPE_CHECKING:
    _: StateHandler = handle_open


async def _disconnect_receiver(state: StateOpen) -> None:
    excs: List[BaseException] = []

    try:
        await state.internal_receiver.unregister_receiver(state.my_receiver_id)
    except BaseException as e:
        excs.append(e)

    for topic in state.my_receiver.exact_subscriptions:
        try:
            await state.internal_receiver.decrement_exact(topic)
        except BaseException as e:
            excs.append(e)

    for _, glob in state.my_receiver.glob_subscriptions:
        try:
            await state.internal_receiver.decrement_glob(glob)
        except BaseException as e:
            excs.append(e)

    if excs:
        raise combine_multiple_exceptions(
            "failed to properly disconnect receiver", excs
        )


SendT = Union[WaitingInternalMessage, SimplePendingSendPreFormatted]


def _cleanup(value: SendT) -> None:
    if value.type == WaitingInternalMessageType.SPOOLED_LARGE:
        value.stream.close()


class VoidingDeque(deque[SendT]):
    def append(self, value: SendT, /) -> None:
        _cleanup(value)

    def appendleft(self, value: SendT, /) -> None:
        _cleanup(value)

    def insert(self, i: int, x: SendT, /) -> None:
        _cleanup(x)

    def extend(self, iterable: Iterable[SendT], /) -> None:
        for v in iterable:
            _cleanup(v)

    def extendleft(self, iterable: Iterable[SendT], /) -> None:
        for v in iterable:
            _cleanup(v)

    def __setitem__(
        self,
        key: Union[int, slice, SupportsIndex],
        value: Union[SendT, Iterable[SendT]],
        /,
    ) -> None:
        if isinstance(key, slice):
            for v in cast(Iterable[SendT], value):
                _cleanup(v)
        else:
            _cleanup(cast(SendT, value))

    def __iadd__(self, other: Iterable[SendT], /) -> "VoidingDeque":
        for v in other:
            _cleanup(v)
        return self

    def __add__(self, other: deque[SendT], /) -> "VoidingDeque":
        for v in other:
            _cleanup(v)
        return self



import asyncio

from lonelypss.ws.handlers.open.check_result import CheckResult
from lonelypss.ws.handlers.open.processors.processor import process_any
from lonelypss.ws.state import StateOpen


async def check_process_task(state: StateOpen) -> CheckResult:
    """If the process task is unset but there are unprocessed messages, sets
    the process task. Otherwise, if the process task has completed, handles the result.

    Raises an exception to indicate we should move to the cleanup and disconnect
    process
    """
    if state.process_task is None:
        if state.unprocessed_messages:
            state.process_task = asyncio.create_task(
                process_any(state, state.unprocessed_messages.popleft())
            )
            return CheckResult.RESTART
        return CheckResult.CONTINUE

    if not state.process_task.done():
        return CheckResult.CONTINUE

    finished_task = state.process_task
    state.process_task = None
    finished_task.result()

    if state.unprocessed_messages:
        # save an event loop cycle
        state.process_task = asyncio.create_task(
            process_any(state, state.unprocessed_messages.popleft())
        )

    return CheckResult.RESTART



import asyncio
from typing import cast

from fastapi import WebSocket

from lonelypss.util.websocket_message import WSMessage


def make_websocket_read_task(websocket: WebSocket) -> asyncio.Task[WSMessage]:
    """Creats an asyncio task that provides a better typed version of websocket.receive()"""
    return cast(asyncio.Task[WSMessage], asyncio.create_task(websocket.receive()))



from fastapi import APIRouter, WebSocket

from lonelypss.middleware.config import get_config_from_request
from lonelypss.middleware.ws_receiver import get_ws_receiver_from_request
from lonelypss.ws.handlers.handler import handle_any
from lonelypss.ws.state import State, StateAccepting, StateType

router = APIRouter()


@router.websocket("/v1/websocket")
async def websocket_endpoint(websocket: WebSocket) -> None:
    """Allows sending and receiving notifications over a websocket connection,
    as opposed to the typical way this library is used (HTTP requests). This is
    helpful for the following scenarios:

    - You need to send a large number of notifications, OR
    - You need to receive a large number of notifications, OR
    - You need to receive notifications for a short period of time before unsubscribing, OR
    - You need to receive some notifications, but you cannot accept incoming HTTP requests

    For maximum compatibility with websocket clients, we only communicate
    over the websocket itself (not the http-level header fields).

    ## COMPRESSION

    For notifications (both posted and received) over websockets, this supports
    using zstandard compression. It will either use an embedded dictionary, a
    precomputed dictionary, or a trained dictionary. Under the typical settings, this:

    - Only considers messages that are between 32 and 16384 bytes for training
    - Will train once after 100kb of data is ready, and once more after 10mb of data is ready,
      then will sample 10mb every 24 hours
    - Will only used the trained dictionary on messages that would be used for training

    ## MESSAGES

    messages always begin as follows

    - 2 bytes (F): flags (interpret as big-endian):
        - least significant bit (1): 0 if headers are expanded, 1 if headers are minimal
    - 2 bytes (T): type of message; see below, depends on if it's sent by a subscriber
      or the broadcaster big-endian encoded, unsigned

    EXPANDED HEADERS:
        - 2 bytes (N): number of headers, big-endian encoded, unsigned
        - REPEAT N:
            - 2 bytes (M): length of header name, big-endian encoded, unsigned
            - M bytes: header name, ascii-encoded
            - 2 bytes (L): length of header value, big-endian encoded, unsigned
            - L bytes: header value

    MINIMAL HEADERS:
    the order of the headers are fixed based on the type, in the order documented.
    Given N headers:
    - Repeat N:
        - 2 bytes (L): length of header value, big-endian encoded, unsigned
        - L bytes: header value

    ## Messages Sent to the Broadcaster

    1: Configure:
        configures the broadcasters behavior; may be set at most once and must be
        sent and confirmed before doing anything else if the url is relevant for
        the authorization header

        headers:
        - x-subscriber-nonce: 32 random bytes representing the subscriber's contribution
            to the nonce. The broadcaster will provide its contribution in the response.
        - x-enable-zstd: 1 byte, big-endian, unsigned. 0 to disable zstandard compression,
            1 to indicate the client is willing to receive zstandard compressed messages.
        - x-enable-training: 1 byte, big-endian, unsigned. 0 to indicate the client will not
        accept custom compression dictionaries, 1 to indicate the client may accept them.
        - x-initial-dict: 2 bytes, big-endian, unsigned. 0 to indicate the client does not
        have a specific preset dictionary in mind to use, otherwise, the id of the preset
        dictionary the client thinks is a good fit for this connection

        body:
            none
    2: Subscribe Exact:
        subscribe to an exact topic

        headers:
            - authorization (url: websocket:<nonce>:<ctr>, see below)
            - x-topic: the topic to subscribe to
        body: none
    3: Subscribe Glob:
        subscribe to a glob pattern

        headers:
            - authorization (url: websocket:<nonce>:<ctr>, see below)
            - x-glob: the glob pattern to subscribe to
        body: none
    4: Unsubscribe Exact:
        unsubscribe from an exact topic

        headers:
            - authorization (url: websocket:<nonce>:<ctr>, see below)
            - x-topic: the topic to unsubscribe from
    5: Unsubscribe Glob:
        unsubscribe from a glob pattern

        headers:
            - authorization (url: websocket:<nonce>:<ctr>, see below)
            - x-glob: the glob pattern to unsubscribe from
    6: Notify:
        send a notification within a single websocket message (typically, max 16MB). this
        can be suitable for arbitrary websocket sizes depending on the configuration of the
        broadcaster (e.g., uvicorn and all intermediaries might limit max ws message sizes)

        headers:
            - authorization (url: websocket:<nonce>:<ctr>, see below)
            - x-identifier identifies the notification so the broadcaster can confirm it
            - x-topic is the topic of the notification
            - x-compressor is a big-endian unsigned integer representing one of the previous
            compression methods enabled by the broadcaster
            - x-compressed-length the total length of the compressed body, big-endian, unsigned, max 8 bytes
            - x-decompressed-length the total length of the decompressed body, big-endian, unsigned, max 8 bytes
            - x-compressed-sha512 the sha-512 hash of the compressed content once all parts are concatenated, 64 bytes

        body: the message to send. must have the same hash as the provided hash
    7: Notify Stream:
        send a notification over multiple websocket messages. this is more likely to work on
        typical setups when the notification payload exceeds 16MB.

        headers:
            - authorization (url: websocket:<nonce>:<ctr>, see below)
            - x-identifier identifies the notify whose compressed body is being appended. arbitrary blob, max 64 bytes
            - x-part-id starts at 0 and increments by 1 for each part. interpreted unsigned, big-endian, max 8 bytes
            - x-topic iff x-part-id is 0, the topic of the notification
            - x-compressor iff x-part-id is 0, either 0 for no compression, 1
              for zstandard compression without a custom dictionary, and
              otherwise the id of the compressor from one of the
              "Enable X compression" broadcaster->subscriber messages
            - x-compressed-length iff x-part-id is 0, the total length of the compressed body, big-endian, unsigned, max 8 bytes
            - x-decompressed-length iff x-part-id is 0, the total length of the decompressed body, big-endian, unsigned, max 8 bytes
            - x-compressed-sha512 iff x-part-id is 0, the sha-512 hash of the compressed content once all parts are concatenated, 64 bytes

        body:
            - blob of data to append to the compressed notification body
    8: Continue Receive:
        confirms that the subscriber received part of a streamed notification and needs more

        headers:
        - x-identifier: the identifier of the notification the subscriber needs more parts for
        - x-part-id: the part id that they received up to, big-endian, unsigned, max 8 bytes

        body: none
    9. Confirm Receive:
        confirms that the subscriber received a streamed notification

        headers:
        - x-identifier: the identifier of the notification that was sent

        body: none

    ## Messages Sent to the Subscriber

    1: Configure Confirmation:
        confirms we received the configuration options from the subscriber

        headers:
            - `x-broadcaster-nonce`: (32 bytes)
                the broadcasters contribution for random bytes to the nonce.
                the connection nonce is SHA256(subscriber_nonce CONCAT broadcaster_nonce),
                which is used in the url for generating the authorization header
                when the broadcaster sends a notification to the receiver over
                this websocket and when the subscriber subscribers to a topic over
                this websocket.

                the url is of the form `websocket:<nonce>:<ctr>`, where the ctr is
                a signed 8-byte integer that starts at 1 (or -1) and that depends on if it
                was sent by the broadcaster or subscriber. Both the subscriber and
                broadcaster keep track of both counters; the subscribers counter
                is always negative and decremented by 1 after each subscribe or unsubscribe
                request, the broadcasters counter is always positive and incremented by 1 after
                each notification sent. The nonce is base64url encoded, the ctr is
                hex encoded without a leading 0x and unpadded, e.g.,
                `websocket:abc123:10ffffffffffffff` or `websocket:abc123:-1a`. note that
                the counter changes every time an authorization header is provided,
                even within a single "operation", so e.g. a Notify Stream message broken
                into 6 parts will change the counter 6 times.

    2: Subscribe Exact Confirmation:
        confirms that the subscriber will receive notifications for the given topic

        headers:
            - x-topic: the topic that the subscriber is now subscribed to

        body: none
    3. Subscribe Glob Confirmation:
        confirms that the subscriber will receive notifications for the given glob pattern

        headers:
            - x-glob: the pattern that the subscriber is now subscribed to

        body: none
    4: Unsubscribe Exact Confirmation:
        confirms that the subscriber will no longer receive notifications for the given topic

        headers:
            - x-topic: the topic that the subscriber is now unsubscribed from

        body: none
    5: Unsubscribe Glob Confirmation:
        confirms that the subscriber will no longer receive notifications for the given glob pattern

        headers:
            - x-glob: the pattern that the subscriber is now unsubscribed from

        body: none
    6: Notify Confirmation:
        confirms that we sent a notification to subscribers; this is also sent
        for streamed notifications after the last part was received by the broadcaster

        headers:
            - x-identifier: the identifier of the notification that was sent
            - x-subscribers: the number of subscribers that received the notification

        body: none
    7: Notify Continue:
        confirms that we received a part of a streamed notification but need more. You
        do not need to wait for this before continuing, and should never retry WS messages
        as the underlying protocol already handles retries. to abort a send, close the WS
        and reconnect

        headers:
            - x-identifier: the identifier of the notification we need more parts for
            - x-part-id: the part id that we received up to, big-endian, unsigned

        body: none
    8: Receive Stream
        tells the subscriber about a notification on a topic they are subscribed to, possibly
        over multiple messages

        headers:
            - authorization (url: websocket:<nonce>:<ctr>, see above)
            - x-identifier identifies the notify whose compressed body is being appended. arbitrary blob, max 64 bytes
            - x-part-id starts at 0 and increments by 1 for each part. interpreted unsigned, big-endian, max 8 bytes
            - x-topic iff x-part-id is 0, the topic of the notification
            - x-compressor iff x-part-id is 0, either 0 for no compression, 1 for no custom dictionary zstd, and
              otherwise the id of the compressor from one of
              the "Enable X compression" broadcaster->subscriber messages
            - x-compressed-length iff x-part-id is 0, the total length of the compressed body, big-endian, unsigned, max 8 bytes
            - x-decompressed-length iff x-part-id is 0, the total length of the decompressed body, big-endian, unsigned, max 8 bytes
            - x-compressed-sha512 iff x-part-id is 0, the sha-512 hash of the compressed content once all parts are concatenated, 64 bytes

        body:
            - blob of data to append to the compressed notification body
    9: Enable zstandard compression with preset dictionary
        configures the subscriber to expect and use a dictionary that it already has available.
        this may use precomputed dictionaries that were specified during the broadcaster's
        configuration with the assumption the subscriber has them

        headers:
            x-identifier: which compressor is enabled, unsigned, big-endian, max 2 bytes, min 1.
                A value of 1 means compression without a custom dictionary.
            x-compression-level: what compression level we think is best when using
                this dictionary. signed, big-endian, max 2 bytes, max 22. the subscriber
                is free to choose a different compression level
            x-min-size: 4 bytes, big-endian, unsigned. a hint to the client for the smallest
                payload for which we think this dictionary is useful. the client can use this
                dictionary on smaller messages if it wants
            x-max-size: 8 bytes, big-endian, unsigned. a hint to the client for the largest
                payload for which we think this dictionary is useful. uses 2**64-1 to indicate
                no upper bound. the client can use this dictionary on larger messages if it wants

        body: none
    10: Enable zstandard compression with a custom dictionary
        configures the subscriber to use a dictionary we just trained

        headers:
            x-identifier: the id we are assigning to this dictionary, unsigned, big-endian, max 8 bytes,
                min 65536. if not unique, disconnect
            x-compression-level: what compression level we think is best when using
                this dictionary. signed, big-endian, max 2 bytes, max 22. the subscriber
                is free to choose a different compression level
            x-min-size: inclusive, max 4 bytes, big-endian, unsigned. a hint to the subscriber for the smallest
                payload for which we think this dictionary is useful. the subscriber can use this
                dictionary on smaller messages if it wants
            x-max-size: exclusive, max 8 bytes, big-endian, unsigned. a hint to the subscriber for the largest
                payload for which the broadcaster will compress with this dictionary. uses 2**64-1 to indicate
                no upper bound. the subscriber can use this dictionary on larger messages if it wants

        body: the dictionary, typically 16-64kb

    11. Disable zstandard compression with a specific custom dictionary
        configures the subscriber to stop using a dictionary the broadcaster previously trained
        and transmitted, and indicates the broadcaster will not use it in the future
        (i.e., the subscriber can clear it from memory)

        headers:
            x-identifier: the id of the dictionary to stop using

        body: none
    """
    config = get_config_from_request(websocket)
    receiver = get_ws_receiver_from_request(websocket)

    state: State = StateAccepting(
        type=StateType.ACCEPTING,
        websocket=websocket,
        broadcaster_config=config,
        internal_receiver=receiver,
    )
    while state.type != StateType.CLOSED:
        state = await handle_any(state)



import asyncio
import base64
import time
from contextlib import asynccontextmanager
from typing import AsyncGenerator

import aiohttp
from lonelypsp.util.cancel_and_check import cancel_and_check

from lonelypss.config.config import Config, MutableMissedInfo


async def sweep_missed_once(config: Config) -> None:
    """
    Sweeps the unsent missed messages within the database one time, without
    looping. Generally, `sweep_missed_forever` or `sweep_missed` (async context manager)
    to have the looping handled as well

    This is concurrency-safe, i.e., there may be multiple concurrently running asyncio
    coroutines or this may be running in multiple processes connected to the same db
    without unexpected behavior (e.g., duplicate MISSED messages or not properly
    incrementing attempts) given reasonable assumptions (locks are not stolen due to
    timeouts)
    """

    async with aiohttp.ClientSession(
        timeout=aiohttp.ClientTimeout(
            total=config.outgoing_http_timeout_total,
            connect=config.outgoing_http_timeout_connect,
            sock_read=config.outgoing_http_timeout_sock_read,
            sock_connect=config.outgoing_http_timeout_sock_connect,
        )
    ) as session:
        async for msg in config.get_overdue_missed_with_lock(now=time.time()):
            if msg.info.subscriber_info.recovery is None:
                if (res := await msg.release(new_info=None)) != "ok":
                    return
                continue

            try:
                authorization = await config.authorize_missed(
                    recovery=msg.info.subscriber_info.recovery,
                    topic=msg.info.topic,
                    now=time.time(),
                )
                try:
                    async with session.post(
                        msg.info.subscriber_info.recovery,
                        headers={
                            **(
                                {"Authorization": authorization}
                                if authorization is not None
                                else {}
                            ),
                            "X-Topic": base64.b64encode(msg.info.topic).decode("ascii"),
                        },
                    ) as resp:
                        if resp.status >= 400 and resp.status < 500:
                            content_type = resp.headers.get("Content-Type")
                            if content_type is not None and content_type.startswith(
                                "application/json"
                            ):
                                content = await resp.json()
                                if (
                                    isinstance(content, dict)
                                    and content.get("unsubscribe") is True
                                ):
                                    if (
                                        res := await msg.release(new_info=None)
                                    ) != "ok":
                                        return
                                    continue
                        resp.raise_for_status()
                except aiohttp.ClientError:
                    next_attempt = msg.info.attempts + 1
                    next_retry_at = await config.get_delay_for_next_missed_retry(
                        receive_url=msg.info.subscriber_info.url,
                        missed_url=msg.info.subscriber_info.recovery,
                        topic=msg.info.topic,
                        attempts=next_attempt,
                    )
                    if next_retry_at is None:
                        if (res := await msg.release(new_info=None)) != "ok":
                            return
                        continue
                    res = await msg.release(
                        new_info=MutableMissedInfo(
                            attempts=next_attempt,
                            next_retry_at=next_retry_at,
                        )
                    )
                    if res != "ok":
                        return
                    continue

                await msg.release(new_info=None)
            except BaseException:
                await msg.release(
                    new_info=MutableMissedInfo(
                        attempts=msg.info.attempts,
                        next_retry_at=msg.info.next_retry_at,
                    )
                )
                raise


async def sweep_missed_forever(config: Config) -> None:
    """
    Sweeps the unsent missed messages within the database forever, looping
    until asyncio.CancelledError is raised
    """

    while True:
        await sweep_missed_once(config)
        await asyncio.sleep(config.sweep_missed_interval)


@asynccontextmanager
async def sweep_missed(config: Config) -> AsyncGenerator[None, None]:
    """
    Context manager for sweeping missed messages within the database until
    the context is exited without interrupting mid-sweep
    """

    exit_event = asyncio.Event()

    async def _inner() -> None:
        while not exit_event.is_set():
            await sweep_missed_once(config)

            want_exit = asyncio.create_task(exit_event.wait())
            want_loop = asyncio.create_task(asyncio.sleep(config.sweep_missed_interval))
            await asyncio.wait(
                [want_exit, want_loop], return_when=asyncio.FIRST_COMPLETED
            )
            await cancel_and_check(want_exit)
            await cancel_and_check(want_loop)

    inner_task = asyncio.create_task(_inner())
    try:
        yield
    finally:
        exit_event.set()
        await inner_task



import time
from typing import TYPE_CHECKING

from lonelypsp.stateful.constants import BroadcasterToSubscriberStatefulMessageType
from lonelypsp.stateful.messages.confirm_subscribe import (
    B2S_ConfirmSubscribeExact,
    serialize_b2s_confirm_subscribe_exact,
)
from lonelypsp.stateful.messages.subscribe import S2B_SubscribeExact

from lonelypss.ws.handlers.open.errors import AuthRejectedException
from lonelypss.ws.handlers.open.processors.protocol import S2B_MessageProcessor
from lonelypss.ws.handlers.open.send_simple_asap import send_simple_asap
from lonelypss.ws.handlers.open.websocket_url import (
    make_for_receive_websocket_url_and_change_counter,
)
from lonelypss.ws.state import StateOpen


async def process_subscribe_exact(
    state: StateOpen, message: S2B_SubscribeExact
) -> None:
    """Processes a request by the subscriber to subscribe to a specific topic,
    receiving notifications within this websocket
    """
    url = make_for_receive_websocket_url_and_change_counter(state)
    auth_at = time.time()
    auth_result = await state.broadcaster_config.is_subscribe_exact_allowed(
        url=url,
        recovery=None,
        exact=message.topic,
        now=auth_at,
        authorization=message.authorization,
    )
    if auth_result != "ok":
        raise AuthRejectedException(f"subscribe exact: {auth_result}")

    if message.topic in state.my_receiver.exact_subscriptions:
        raise Exception("already subscribed to exact topic")

    # note we confirm before registering to ensure they don't receive notifications
    # on the topic before its been confirmed
    send_simple_asap(
        state,
        serialize_b2s_confirm_subscribe_exact(
            B2S_ConfirmSubscribeExact(
                type=BroadcasterToSubscriberStatefulMessageType.CONFIRM_SUBSCRIBE_EXACT,
                topic=message.topic,
            ),
            minimal_headers=state.broadcaster_config.websocket_minimal_headers,
        ),
    )
    state.my_receiver.exact_subscriptions.add(message.topic)
    await state.internal_receiver.increment_exact(message.topic)


if TYPE_CHECKING:
    _: S2B_MessageProcessor[S2B_SubscribeExact] = process_subscribe_exact



import asyncio
import io
from typing import cast

from lonelypsp.stateful.constants import SubscriberToBroadcasterStatefulMessageType
from lonelypsp.stateful.parser import S2B_AnyMessageParser
from lonelypsp.stateful.parser_helpers import parse_s2b_message_prefix

from lonelypss.util.websocket_message import WSMessageBytes
from lonelypss.ws.handlers.open.check_result import CheckResult
from lonelypss.ws.handlers.open.errors import NormalDisconnectException
from lonelypss.ws.handlers.open.processors.processor import process_any
from lonelypss.ws.state import (
    StateOpen,
)
from lonelypss.ws.util import make_websocket_read_task


async def check_read_task(state: StateOpen) -> CheckResult:
    """Makes progress using the result of the read task, if possible. Raises
    an exception to indicate that we should begin the cleanup and shutdown
    process
    """
    if not state.read_task.done():
        return CheckResult.CONTINUE

    result = state.read_task.result()
    if result["type"] == "websocket.disconnect":
        raise NormalDisconnectException

    if "bytes" not in result:
        raise Exception("unexpected message type (expected bytes)")

    payload = cast(WSMessageBytes, result)["bytes"]
    payload_reader = io.BytesIO(payload)
    prefix = parse_s2b_message_prefix(payload_reader)
    if prefix.type == SubscriberToBroadcasterStatefulMessageType.CONFIGURE:
        raise ValueError("already configured")

    message = S2B_AnyMessageParser.parse(prefix.flags, prefix.type, payload_reader)
    state.read_task = make_websocket_read_task(state.websocket)

    # fast track acks as they are common and can be handled synchronously
    if message.type == SubscriberToBroadcasterStatefulMessageType.CONFIRM_RECEIVE:
        expected_ack = state.expecting_acks.get_nowait()
        if expected_ack.type != message.type:
            raise Exception(
                f"unexpected confirm receive (expecting a {expected_ack.type})"
            )
        if expected_ack.identifier != message.identifier:
            raise Exception(
                f"unexpected confirm receive (expecting identifier {expected_ack.identifier!r}, got {message.identifier!r})"
            )
        return CheckResult.RESTART

    if message.type == SubscriberToBroadcasterStatefulMessageType.CONTINUE_RECEIVE:
        expected_ack = state.expecting_acks.get_nowait()
        if expected_ack.type != message.type:
            raise Exception(
                f"unexpected continue receive (expecting a {expected_ack.type})"
            )
        if expected_ack.identifier != message.identifier:
            raise Exception(
                f"unexpected continue receive (expecting identifier {expected_ack.identifier!r}, got {message.identifier!r})"
            )
        if expected_ack.part_id != message.part_id:
            raise Exception(
                f"unexpected continue receive (expecting part_id {expected_ack.part_id}, got {message.part_id})"
            )
        return CheckResult.RESTART

    if state.process_task is None:
        state.process_task = asyncio.create_task(process_any(state, message))
    else:
        state.unprocessed_messages.append(message)

    return CheckResult.RESTART



from typing import TYPE_CHECKING

from fastapi.websockets import WebSocketState

from lonelypss.ws.handlers.protocol import StateHandler
from lonelypss.ws.state import State, StateClosed, StateType


async def handle_closing(state: State) -> State:
    """Closes the websocket normally then raises the indicated exception, if any,
    otherwise moves to CLOSED
    """

    assert state.type == StateType.CLOSING
    if state.websocket.client_state != WebSocketState.DISCONNECTED:
        await state.websocket.close()
    if state.exception is not None:
        raise state.exception
    return StateClosed(type=StateType.CLOSED)


if TYPE_CHECKING:
    _: StateHandler = handle_closing



from lonelypsp.stateful.constants import BroadcasterToSubscriberStatefulMessageType
from lonelypsp.stateful.messages.enable_zstd_custom import (
    B2S_EnableZstdCustom,
    serialize_b2s_enable_zstd_custom,
)
from lonelypsp.stateful.messages.enable_zstd_preset import (
    B2S_EnableZstdPreset,
    serialize_b2s_enable_zstd_preset,
)

from lonelypss.ws.handlers.open.check_result import CheckResult
from lonelypss.ws.handlers.open.send_simple_asap import send_simple_asap
from lonelypss.ws.state import CompressorState, StateOpen


async def check_compressors(state: StateOpen) -> CheckResult:
    """Checks to see if any of the compressors are in the preparing state
    but the preparation is finished

    Raises an exception to indicate we should move to the cleanup and disconnect
    process
    """

    did_something = False
    for idx in range(len(state.compressors)):
        compressor = state.compressors[idx]
        if compressor.type == CompressorState.PREPARING and compressor.task.done():
            new_compressor = compressor.task.result()
            state.compressors[idx] = new_compressor

            if new_compressor.identifier < 65536:
                send_simple_asap(
                    state,
                    serialize_b2s_enable_zstd_preset(
                        B2S_EnableZstdPreset(
                            type=BroadcasterToSubscriberStatefulMessageType.ENABLE_ZSTD_PRESET,
                            identifier=new_compressor.identifier,
                            compression_level=new_compressor.level,
                            min_size=(
                                state.broadcaster_config.compression_min_size
                                if new_compressor.identifier != 1
                                else state.broadcaster_config.compression_trained_max_size
                            ),
                            max_size=2**64 - 1,
                        ),
                        minimal_headers=state.broadcaster_config.websocket_minimal_headers,
                    ),
                )
            else:
                assert (
                    new_compressor.data is not None
                ), f"compressor identifier {compressor.identifier}>=2**16 must have data"
                send_simple_asap(
                    state,
                    serialize_b2s_enable_zstd_custom(
                        B2S_EnableZstdCustom(
                            type=BroadcasterToSubscriberStatefulMessageType.ENABLE_ZSTD_CUSTOM,
                            identifier=new_compressor.identifier,
                            compression_level=new_compressor.level,
                            min_size=state.broadcaster_config.compression_min_size,
                            max_size=state.broadcaster_config.compression_trained_max_size,
                            dictionary=new_compressor.data.as_bytes(),
                        ),
                        minimal_headers=state.broadcaster_config.websocket_minimal_headers,
                    ),
                )

            did_something = True

    return CheckResult.RESTART if did_something else CheckResult.CONTINUE



import re
import time
from typing import TYPE_CHECKING

from lonelypsp.stateful.constants import BroadcasterToSubscriberStatefulMessageType
from lonelypsp.stateful.messages.confirm_subscribe import (
    B2S_ConfirmSubscribeGlob,
    serialize_b2s_confirm_subscribe_glob,
)
from lonelypsp.stateful.messages.subscribe import S2B_SubscribeGlob

from lonelypss.ws.handlers.open.errors import AuthRejectedException
from lonelypss.ws.handlers.open.processors.protocol import S2B_MessageProcessor
from lonelypss.ws.handlers.open.send_simple_asap import send_simple_asap
from lonelypss.ws.handlers.open.websocket_url import (
    make_for_receive_websocket_url_and_change_counter,
)
from lonelypss.ws.state import StateOpen

try:
    from glob import translate as _glob_translate  # type: ignore

    def translate(pat: str) -> str:
        return _glob_translate(pat, recursive=True, include_hidden=True)

except ImportError:
    from fnmatch import translate


async def process_subscribe_glob(state: StateOpen, message: S2B_SubscribeGlob) -> None:
    """Processes a request by the subscriber to subscribe to utf-8 decodable
    topics which match the given glob pattern, receiving notifications within
    this websocket
    """
    url = make_for_receive_websocket_url_and_change_counter(state)
    auth_at = time.time()
    auth_result = await state.broadcaster_config.is_subscribe_glob_allowed(
        url=url,
        recovery=None,
        glob=message.glob,
        now=auth_at,
        authorization=message.authorization,
    )
    if auth_result != "ok":
        raise AuthRejectedException(f"subscribe exact: {auth_result}")

    if any(message.glob == glob for _, glob in state.my_receiver.glob_subscriptions):
        raise Exception("already subscribed to glob pattern")

    glob_regex = re.compile(translate(message.glob))

    # note we confirm before registering to ensure they don't receive notifications
    # on the topic before its been confirmed
    send_simple_asap(
        state,
        serialize_b2s_confirm_subscribe_glob(
            B2S_ConfirmSubscribeGlob(
                type=BroadcasterToSubscriberStatefulMessageType.CONFIRM_SUBSCRIBE_GLOB,
                glob=message.glob,
            ),
            minimal_headers=state.broadcaster_config.websocket_minimal_headers,
        ),
    )
    state.my_receiver.glob_subscriptions.append((glob_regex, message.glob))
    await state.internal_receiver.increment_glob(message.glob)


if TYPE_CHECKING:
    _: S2B_MessageProcessor[S2B_SubscribeGlob] = process_subscribe_glob
